#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Glob User Config
=================

Substitute globular expressions in `config.yaml` for entries
within the local registry


Contents
========

Functions
-------

    glob_read_write - glob expressions in 'read' and 'write' blocks

"""

__date__ = "2021-08-16"

import typing
import copy

import fair.registry.requests as fdp_req
import fair.configuration as fdp_conf
import fair.exceptions as fdp_exc
import fair.registry.versioning as fdp_ver
import fair.utilities as fdp_util


def glob_read_write(
    user_config: typing.Dict,
    blocktype: str,
    version: str,
    registry_url: str,
    search_key: str = None,
    remove_wildcard: bool = False) -> typing.List:
    """Substitute glob expressions in the 'read' or 'write' part of a user config

    Parameters
    ----------
    user_config : typing.Dict
        config yaml
    blocktype : str
        block type to process
    version : str
        version string
    registry_url : str
        URL of the registry to process
    search_key : str, optional
        key to search under, default is taken from SEARCH_KEYS
    remove_wildcard: bool, optional
        whether to delete wildcard from yaml file, default is False
    """
    
    _block_cfg = user_config[blocktype]
    _parsed: typing.List[typing.Dict] = []

    # Iterate through all entries in the section looking for any
    # key-value pairs that contain glob statements.
    for entry in _block_cfg:
        # We still want to keep the wildcard version in case the
        # user wants to write to this namespace.
        # Wipe version info for this object to start from beginning
        _orig_entry = copy.deepcopy(entry)

        _orig_entry['use']['version'] = str(fdp_ver.get_correct_version(version, free_write=blocktype!='read'))

        _glob_vals = [(k, v) for k, v in entry.items() if isinstance(v, str) and '*' in v]
        if len(_glob_vals) > 1:
            # For now only allow one value within the dictionary to have them
            raise fdp_exc.NotImplementedError(
                "Only one key-value pair in a 'read' list entry may contain a"
                " globbable value"
            )
        elif len(_glob_vals) == 0:
            # If no globbables keep existing statement
            _parsed.append(entry)
            continue
        elif not remove_wildcard:
            # If we're going ahead, add in the wildcard too if wanted
            _parsed.append(_orig_entry)

        _key_glob, _globbable = _glob_vals[0]

        if not search_key:
            search_key = fdp_req.SEARCH_KEYS[_key_glob]

        _search_dict = {search_key: _globbable}

        # Update search from 'use' block
        _search_dict.update(entry['use'])

        try:
            fdp_ver.parse_incrementer(_search_dict["version"])
            # Is an incrementer, so get rid of it
            _search_dict.pop('version', None)
        except fdp_exc.UserConfigError: # Should be an exact version, so keep
            None

        # Send a request to the relevant registry using the search string
        # and the selected search key        
        _results = fdp_req.get(
            registry_url,
            _key_glob,
            params = _search_dict
        )

        # Iterate through all results, make a copy of the entry and swap
        # the globbable statement for the result statement appending this
        # to the output list
        for result in _results:
            _entry_dict = copy.deepcopy(entry)
            if _key_glob in _entry_dict['use']:
                _entry_dict['use'][_key_glob] = result[search_key]
            if _key_glob in _entry_dict:
                _entry_dict[_key_glob] = result[search_key]
            _parsed.append(_entry_dict)

    # Before returning the list of dictionaries remove any duplicates
    user_config[blocktype] = fdp_util.remove_dictlist_dupes(_parsed)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Execute Job
===========

Creates required files for job execution and runs the job itself using the
retrieved metadata

"""

__date__ = "2021-06-30"

import os
import sys
import platform
import typing
import glob
import logging
import hashlib
import datetime
import subprocess
import click

import fair.configuration as fdp_conf
import fair.common as fdp_com
import fair.history as fdp_hist
import fair.exceptions as fdp_exc
import fair.user_config as fdp_user

from fair.common import CMD_MODE

logger = logging.getLogger("FAIRDataPipeline.Run")

# Dictionary of recognised shell labels.
SHELLS: typing.Dict[str, str] = {
    "pwsh": {
        "exec": "pwsh -command \". '{0}'\"",
        "extension": "ps1"
    },
    "batch": {
        "exec": "{0}",
        "extension": "bat"
    },
    "powershell": {
        "exec": "powershell -command \". '{0}'\"",
        "extension": "ps1"
    },
    "python2": {
        "exec": "python2 {0}",
        "extension": "py"
    },
    "python3": {
        "exec": "python3 {0}",
        "extension": "py"
    },
    "python": {
        "exec": "python {0}",
        "extension": "py"
    },
    "R": {
        "exec": "R -f {0}",
        "extension": "R"
    },
    "julia": {
        "exec": "julia {0}",
        "extension": "jl"
    },
    "bash": {
        "exec": "bash -eo pipefail {0}",
        "extension": "sh",
    },
    "java": {
        "exec": "java {0}",
        "extension": "java"
    },
    "sh": {
        "exec":"sh -e {0}",
        "extension": "sh"
    }
}


def run_command(
    repo_dir: str,
    config_yaml: str = None,
    mode: CMD_MODE = CMD_MODE.RUN,
    bash_cmd: str = "",
) -> str:
    """Execute a process as part of job

    Executes a command from the given job config file, if a command is
    given this is job instead and overwrites that in the job config file.

    Parameters
    ----------
    local_uri : str
        local registry endpoint
    repo_dir : str
        directory of repository to run from
    config_yaml : str, optional
        run from a given config.yaml file
    bash_cmd : str, optional
        override execution command with a bash command
    """

    if not config_yaml:
        config_yaml = os.path.join(fdp_com.find_fair_root(), fdp_com.USER_CONFIG_FILE)

    logger.debug("Using user configuration file: %s", config_yaml)
    click.echo(f"Updating registry from {config_yaml}", err = True)

    # Record the time the job was commenced, create a log and both
    # print output and write it to the log file
    _now = datetime.datetime.now()
    _timestamp = _now.strftime("%Y-%m-%d_%H_%M_%S_%f")
    _logs_dir = fdp_hist.history_directory(repo_dir)

    if not os.path.exists(_logs_dir):
        os.mkdir(_logs_dir)

    _log_file = os.path.join(_logs_dir, f"job_{_timestamp}.log")

    # Check that the specified user config file for a job actually exists
    if not os.path.exists(config_yaml):
        raise fdp_exc.FileNotFoundError(
            "Failed to read user configuration, "
            f"file '{config_yaml}' does not exist."
        )

    logger.debug(f"Creating user configuration job object from {config_yaml}")

    _job_cfg = fdp_user.JobConfiguration(config_yaml)

    _job_cfg.update_from_fair(repo_dir)

    if bash_cmd:
        _job_cfg.set_command(bash_cmd)

    _job_dir = os.path.join(fdp_com.default_jobs_dir(), _timestamp)
    logger.debug("Using job directory: %s", _job_dir)
    os.makedirs(_job_dir, exist_ok=True)

    _job_cfg.prepare(_job_dir, _timestamp, mode)

    _run_executable = "script" in _job_cfg["run_metadata"] or "script_path" in _job_cfg["run_metadata"]
    _run_executable = _run_executable and mode in [CMD_MODE.RUN, CMD_MODE.PASS]

    if mode == CMD_MODE.PASS:
        logger.debug(
            "Run called in passive mode, no command will be executed"
        )

    # Set location of working config.yaml to the job directory
    _work_cfg_yml = os.path.join(_job_dir, fdp_com.USER_CONFIG_FILE)

    # Fetch the CLI configurations for logging information
    _user = fdp_conf.get_current_user_name(repo_dir)
    _email = fdp_conf.get_current_user_email(repo_dir)

    if mode in [CMD_MODE.PULL]:
        # If not a fair run then the log file will have less metadata
        # all commands should produce a log so that the 'fair log' history
        # can be displayed
        with open(_log_file, "a") as f:
            _out_str = _now.strftime("%a %b %d %H:%M:%S %Y %Z")
            f.writelines(
                [
                    "--------------------------------\n",
                    f" Commenced = {_out_str}\n",
                    f" Author    = {' '.join(_user)} <{_email}>\n",
                    ' Command   = fair pull\n',
                    "--------------------------------\n",
                ]
            )

    _job_cfg.write(_work_cfg_yml)

    logger.debug("Creating working configuration storage location")

    if _run_executable:

        # Create a run script if 'script' is specified instead of 'script_path'
        # else use the script
        _cmd_setup = setup_job_script(
            _job_cfg.content,
            _job_cfg.env['FDP_CONFIG_DIR'],
            _job_dir
        )

        _job_cfg.set_script(_cmd_setup['script'])
        _job_cfg.write(_work_cfg_yml)

        if _job_cfg.shell not in SHELLS:
            raise fdp_exc.UserConfigError(
                f"Unrecognised shell '{_job_cfg.shell}' specified."
            )

        _exec = SHELLS[_job_cfg.shell]["exec"]
        _cmd_list = _exec.format(_cmd_setup['script']).split()

        if not _job_cfg.command:
            click.echo("Nothing to run.")
            sys.exit(0)

        # Generate a local job log for the CLI, this is NOT
        # related to metadata sent to the registry
        # this log is viewable via the `fair view <run-cli-sha>`
        with open(_log_file, "a") as f:
            _out_str = _now.strftime("%a %b %d %H:%M:%S %Y %Z")
            _user = _user[0] if not _user[1] else ' '.join(_user)
            f.writelines(
                [
                    "--------------------------------\n",
                    f" Commenced = {_out_str}\n",
                    f" Author    = {_user} <{_email}>\n",
                    f" Namespace = {_job_cfg.default_output_namespace}\n",
                    f" Command   = {' '.join(_cmd_list)}\n",
                    "--------------------------------\n",
                ]
            )

        if mode == CMD_MODE.RUN:
            execute_run(_cmd_list, _job_cfg, _log_file, _now)
        else: # CMD_MODE.PASS
            _end_time = datetime.datetime.now()
            with open(_log_file, "a") as f:
                _duration = _end_time - _now
                f.writelines(
                    [
                        "Operating in ci mode without running script\n",
                        f"------- time taken {_duration} -------\n"
                    ]
                )
    else:
        _end_time = datetime.datetime.now()
        with open(_log_file, "a") as f:
            _duration = _end_time - _now
            f.writelines([f"------- time taken {_duration} -------\n"])


    return get_job_hash(_job_dir)


def execute_run(
    command: typing.List[str],
    job_config: fdp_user.JobConfiguration,
    log_file: str,
    timestamp: datetime.datetime) -> None:
    """Execute a run initialised by a CLI run of mode RUN

    Parameters
    ----------
    command : typing.List[str]
        command to execute
    job_config : str
        job configuration
    log_file : str
        log file to write stdout
    timestamp : datetime.datetime
        job execution start time

    Raises
    ------
    fdp_exc.CommandExecutionError
        [description]
    """
    logger.debug("Executing command: %s", ' '.join(command))

    # Run the submission script
    _process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        bufsize=1,
        text=True,
        shell=False,
        env=job_config.environment,
        cwd=job_config.local_repository
    )

    # Write any stdout to the job log
    for line in iter(_process.stdout.readline, ""):
        with open(log_file, "a") as f:
            f.writelines([line])
        click.echo(line, nl=False)
        sys.stdout.flush()
    _process.wait()
    _end_time = datetime.datetime.now()
    with open(log_file, "a") as f:
        _duration = _end_time - timestamp
        f.writelines([f"------- time taken {_duration} -------\n"])

    # Exit the session if the job failed
    if _process.returncode != 0:
        raise fdp_exc.CommandExecutionError(
            f"Run failed with exit code '{_process.returncode}'",
            exit_code=_process.returncode
        )


def get_job_hash(job_dir: str) -> str:
    """Retrieve the hash for a given job

    NOTE: A job can consist of multiple code runs if the API implementation
    called initiates multiple executions. "Job" here refers to a call of
    'fair run'.

    Parameters
    ----------
    job_dir : str
        jobs directory

    Returns
    -------
    str
        hash of job
    """
    if not os.path.exists(job_dir):
        raise fdp_exc.FileNotFoundError(
            "Failed to find hash for job, "
            f"directory '{job_dir}' does not exist."
        )
    _directory = os.path.abspath(job_dir)
    return hashlib.sha1(_directory.encode("utf-8")).hexdigest()


def get_job_dir(job_hash: str) -> str:
    """Get job directory from a hash

    Parameters
    ----------
    job_hash : str
        hash for a given job

    Returns
    -------
    str
        associated job directory
    """
    _jobs = glob.glob(os.path.join(fdp_com.default_jobs_dir(), '*'))

    for job in _jobs:
        _hash = hashlib.sha1(os.path.abspath(job).encode("utf-8")).hexdigest()
        if _hash == job_hash:
            return job

    return ""


def setup_job_script(
    user_config: typing.Dict,
    config_dir: str,
    output_dir: str
    ) -> typing.Dict[str, typing.Any]:
    """Setup a job script from the given configuration.

    Checks the user configuration file for the required 'script' or 'script_path'
    keys and determines the process to be executed. Also sets up an environment
    usable when executing the submission script.

    Parameters
    ----------
    local_repo : str
        local FAIR repository
    script : str
        script to write to file
    config_dir : str
        final location of output config.yaml
    output_dir : str
        location to store submission/job script

    Returns
    -------
    Dict[str, Any]
        a dictionary containing information on the command to execute,
        which shell to run it in and the environment to use
    """
    logger.debug("Setting up job script for execution")
    _cmd = None

    if config_dir[-1] != os.path.sep:
        config_dir += os.path.sep

    # Check if a specific shell has been defined for the script
    _shell = None
    _out_file = None

    if "shell" in user_config["run_metadata"]:
        _shell = user_config["run_metadata"]["shell"]
    else:
        _shell = "batch" if platform.system() == "Windows" else "sh"

    logger.debug("Will use shell: %s", _shell)

    if "script" in user_config["run_metadata"]:
        _cmd = user_config["run_metadata"]['script']

        if 'extension' not in SHELLS[_shell]:
            raise fdp_exc.InternalError(
                f"Failed to retrieve an extension for shell '{_shell}'"
            )
        _ext = SHELLS[_shell]["extension"]
        _out_file = os.path.join(output_dir, f"script.{_ext}")
        if _cmd:
            with open(_out_file, "w") as f:
                f.write(_cmd)

    elif "script_path" in user_config["run_metadata"]:
        _path = user_config["run_metadata"]["script_path"]
        if not os.path.exists(_path):
            raise fdp_exc.CommandExecutionError(
                f"Failed to execute run, script '{_path}' was not found, or"
                " failed to be created.",
                exit_code=1
            )
        _cmd = open(_path).read()
        _out_file = os.path.join(output_dir, os.path.basename(_path))
        if _cmd:
            with open(_out_file, "w") as f:
                f.write(_cmd)
    
    logger.debug("Script command: %s", _cmd)
    logger.debug("Script written to: %s", _out_file)

    if not _cmd or not _out_file:
        raise fdp_exc.UserConfigError(
            "Configuration file must contain either a valid "
            "'script' or 'script_path' entry under 'run_metadata'"
        )

    return {"shell": _shell, "script": _out_file}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fair', 'fair.metadata', 'fair.parsing', 'fair.registry', 'fair.templates']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.18,<4.0.0',
 'Jinja2>=3.0.1,<4.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'click>=8.0.0,<9.0.0',
 'requests>=2.23.0,<3.0.0',
 'rich>=10.2.3,<11.0.0',
 'semver>=2.13.0,<3.0.0',
 'simplejson>=3.17.5,<4.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['fair = fair.cli:cli']}

setup_kwargs = {
    'name': 'fair-cli',
    'version': '0.2.3',
    'description': 'Synchronization interface for the SCRC FAIR Data Pipeline registry',
    'long_description': '# FAIR Data Pipeline Command Line Interface\n\n[![PyPI](https://img.shields.io/pypi/v/fair-cli)](https://pypi.org/project/fair-cli/) [![PyPI - Python Version](https://img.shields.io/pypi/pyversions/fair-cli)](https://pypi.org/project/fair-cli/) [![DOI](https://zenodo.org/badge/377398464.svg)](https://zenodo.org/badge/latestdoi/377398464) ![PyPI - License](https://img.shields.io/pypi/l/fair-cli)\n\n[![FAIR Data Pipeline CLI](https://github.com/FAIRDataPipeline/FAIR-CLI/actions/workflows/fair-cli.yaml/badge.svg?branch=main)](https://github.com/FAIRDataPipeline/FAIR-CLI/actions/workflows/fair-cli.yaml)  [![codecov](https://codecov.io/gh/FAIRDataPipeline/FAIR-CLI/branch/dev/graph/badge.svg?token=h93TkTiiWf)](https://codecov.io/gh/FAIRDataPipeline/FAIR-CLI) [![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=FAIRDataPipeline_FAIR-CLI&metric=alert_status)](https://sonarcloud.io/summary/new_code?id=FAIRDataPipeline_FAIR-CLI)\n\nFAIR-CLI forms the main interface for synchronising changes between local and shared remote FAIR Data Pipeline registries, it is also used to instantiate model runs/data submissions to the pipeline. Full documentation of the FAIR Data Pipeline can be found on the project [website](https://www.fairdatapipeline.org/).\n\n## Installation\n\nThe package is installed using Pip:\n\n```sh\npip install fair-cli\n```\n\nTo enable tab completion you need to modify your shell:\n\n### Bash\n```\n_FAIR_COMPLETE=bash_source fair > ~/.config/.fair-complete.bash\necho \'. ~/.config/.fair-complete.bash\' >> ~/.bashrc\n```\n\n### zsh\n```\n_FAIR_COMPLETE=zsh_source fair > ~/.fair-complete.zsh\necho \'. ~/.fair-complete.zsh\' >> ~/.bashrc\n```\n\n### Fish\n```\n_FAIR_COMPLETE=bash_source fair > ~/.config/fish/.fair-complete.fish\necho \'. ~/.config/fish/.fair-complete.fish\' >> ~/.bashrc\n```\n\n## Uninstallation\nTo uninstall the CLI run:\n```\nfair purge --all\npip uninstall fair\n```\n\n## The User Configuration File\nJob runs are configured via `config.yaml` files. Upon initialisation of a project, FAIR-CLI automatically generates a starter configuration file with all requirements in place. To execute a process (e.g. perform a model run from a compiled binary/script) an additional key of either `script` or `script_path` must be provided. Alternatively the command `fair run bash` can be used to append the key and run a command directly.\n\nBy default the shell used to execute a process is `sh` or `pwsh` for UNIX and Windows systems respectively. This can be overwritten by assigning the optional `shell` key with one of the following values (where `{0}` is the script file):\n\n| **Shell**    | **Command**                     |\n| ------------ | ------------------------------- |\n| `bash`       | `bash -eo pipefail {0}`         |\n| `java`       | `java {0}`                      |\n| `julia`      | `julia {0}`                     |\n| `powershell` | `powershell -command ". \'{0}\'"` |\n| `pwsh`       | `pwsh -command ". \'{0}\'"`       |\n| `python2`    | `python2 {0}`                   |\n| `python3`    | `python3 {0}`                   |\n| `python`     | `python {0}`                    |\n| `R`          | `R -f {0}`                      |\n| `sh`         | `sh -e {0}`                     |\n\nA full description of `config.yaml` files can be found [here](https://www.fairdatapipeline.org/docs/interface/config/).\n\n## Available Commands\n\n### `init`\n\nInitialises a new FAIR repository within the given directory. This should ideally be the same location as the `.git` folder for the current project, however during setup an option is given to specify an alternative. The command will ask the user a series of questions which will provide metadata for tracking run authors, and also allow for the creation of a starter `config.yaml` file. Initialisation will also configure the CLI itself.\n\n#### Custom CLI Configuration\nAfter setup is complete, the current CLI configuration can also be saved using the command:\n```\nfair init --export\n```\nthe created file can then be re-read at a later point during setup. Alternatively, if creating a configuration from scratch the YAML file should contain the following information:\n\n```yaml\nnamespaces: \n  input: testing\n  output: testing\nregistries:\n  local:\n    data_store: /path/to/local/data_store/,\n    directory: /local/registry/install/directory\n    uri: http://127.0.0.1:8000/api/\n  origin:\n    data_store: /remote/registry/data/store/path/\n    token: /path/to/remote/token\n    uri: https://data.scrc.uk/api/\'\nuser:\n  email: \'test@noreply\',\n  family_name: \'Test\'\n  given_names: \'Interface\'\n  orcid: None,\n  uuid: \'2ddb2358-84bf-43ff-b2aa-3ac7dc3b49f1\'\ngit:\n  local_repo: /local/repo/path\n  remote: origin\ndescription: Testing Project\n```\nthis file is then read during the initialisation:\n\n```sh\nfair init --using <cli-config.yaml file>\n```\n\nFor integration into a CI workflow, the setup can be skipped by running:\n\n```sh\nfair init --ci\n```\n\nwhich will create temporary directories for some of the required location paths.\n\n\n### `run`\n\nThe purpose of `run` is to execute a model/submission run and submit results to the local registry. Outputs of a run will be stored within the `coderun` folder in the directory specified under the `data_store` tag in the `config.yaml`, by default this is `$HOME/.fair/data/coderun`.\n\n```sh\nfair run\n```\n\nIf you wish to use an alternative `config.yaml` then specify it as an additional argument:\n\n```sh\nfair run /path/to/config.yaml\n```\n\nYou can also launch a bash command directly, this will be automatically written into the `config.yaml`:\n\n```sh\nfair run --script "echo \\"Hello World\\""\n```\n\nnote the command itself must be quoted as it is a single argument.\n\n### `pull`\n\nCurrently `pull` will update any entries within the `config.yaml` under the `register` heading creating `external_object` and `data_product` objects on the registry and downloading the data to the local data storage. Any data required for a run is downloaded  and stored within the local registry.\n\n### `purge`\n\nThe `purge` command removes setup of the current project so it can bereinitialised:\n\n```sh\nfair purge\n```\n\nTo remove all configurations entirely (including those global to all projects) run:\n\n```sh\nfair purge --global\n```\n\nFinally the data directory itself can be removed by running:\n\n```sh\nfair purge --data\n```\n\n**WARNING**: This is not recommended as the registry may still have entries pointing to this location!\n\nYou can skip any confirmation messages by running:\n\n```sh\nfair purge --yes\n```\n\n### `registry`\n\nBy default the CLI will launch the registry whenever a synchronisation or run is called. The server will only be halted once all ongoing CLI processes (in the case of multiple parallel calls) have been completed.\n\nHowever the user may also specify a manual launch that will override this behaviour, instead leaving the server running constantly allowing them to view the registry in the browser.\n\nThe commands:\n\n```sh\nfair registry start\n```\n\nand\n\n```sh\nfair registry stop\n```\n\nwill launch and halt the server respectively.\n\n### `log`\n\nRuns are logged locally within the local FAIR repository. A full list of runs is shown by running:\n\n```sh\nfair log\n```\n\nThis will present a list of runs in a summary analogous to a `git log` call:\n\n```yaml\nrun 0db35c20946a1ebeaafdc3b30103cd74a57eb6b6\nAuthor: Joe Bloggs <jbloggs@noreply.uk>\nDate:   Wed Jun 30 09:09:30 2021\n```\n\n| **NOTE**                                                                                                                            |\n| ----------------------------------------------------------------------------------------------------------------------------------- |\n| The SHA for a job is *not* related to a registry code run identifier as multiple code runs can be executed within a single job. |\n\n### `view`\n\nTo view the `stdout` of a run given its SHA as shown by running `fair log` use the command:\n\n```sh\nfair view <sha>\n```\n\nyou do not need to specify the full SHA but rather the first few unique characters.\n\n## Template Variables\n\nWithin the `config.yaml` file, template variables can be specified by using the notation `${{ VAR }}`, the following variables are currently recognised:\n\n| **Variable**        | **Description**                                                                  |\n| ------------------- | -------------------------------------------------------------------------------- |\n| `DATE`              | Date in the form `%Y%m%d`                                                        |\n| `DATETIME`          | Date and time in the form `%Y-%m-%sT%H:%M:S`                                     |\n| `DATETIME-%Y%H%M`   | Date and time in custom format (where `%Y%H%M` can be any valid form)            |\n| `USER`              | The current user as defined in the CLI                                           |\n| `REPO_DIR`          | The FAIR repository root directory                                               |\n| `CONFIG_DIR`        | The directory containing the `config.yaml` after template substitution           |\n| `LOCAL_TOKEN`       | The token for access to the local registry                                       |\n| `SOURCE_CONFIG`     | Path of the user defined `config.yaml`                                           |\n| `GIT_BRANCH`        | Current branch of the `git` repository                                           |\n| `GIT_REMOTE`        | The URI of the git repository specified during setup                             |\n| `GIT_TAG`           | The latest tag on `git`                                                          |\n',
    'author': 'Richard Reeve',
    'author_email': 'richard.reeve@glasgow.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.fairdatapipeline.org/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Command Line Interface
======================

Main command line interface setup script for creation of commands used to
interact with the synchronisation tool.
"""

__date__ = "2021-06-24"

import typing
import os
import sys
import click
import click.shell_completion
import yaml
import pathlib

import fair.session as fdp_session
import fair.common as fdp_com
import fair.history as fdp_hist
import fair.configuration as fdp_conf
import fair.exceptions as fdp_exc
import fair.registry.server as fdp_svr
import fair.run as fdp_run

__author__ = "Scottish COVID Response Consortium"
__credits__ = [
    "Richard Reeve (University of Glasgow)",
    "Nathan Cummings (UKAEA)",
    "Kristian Zarebski (UKAEA)",
    "Dennis Reddyhoff (University of Sheffield)"
]
__license__ = "BSD-2-Clause"
__status__ = "Development"
__copyright__ = "Copyright 2021, FAIR Data Pipeline"

def complete_yamls(ctx, param, incomplete):
    _file_list: typing.List[str] = [
        str(i)
        for i in pathlib.Path(os.getcwd()).rglob('*.yaml')
    ]
    _file_list += [str(i) for i in pathlib.Path(os.getcwd()).rglob('*.yml')]
    return [
        click.shell_completion.CompletionItem(k)
        for k in _file_list if k.startswith(incomplete)
    ]


@click.group()
@click.version_option()
def cli():
    """Welcome to FAIR-CLI, the FAIR data pipeline command-line interface."""
    pass


@cli.command()
@click.option("--verbose/--not-verbose", help="Display URLs", default=False)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def status(verbose, debug) -> None:
    """Get the status of files under staging"""
    try:
        with fdp_session.FAIR(
            os.getcwd(), debug=debug, server_mode=fdp_svr.SwitchMode.CLI
        ) as fair_session:
            fair_session.status(verbose)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.command()
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.argument("output", nargs=-1)
def create(debug, output: str) -> None:
    """Generate a new FAIR repository user YAML config file"""
    output = os.path.join(os.getcwd(), fdp_com.USER_CONFIG_FILE) if not output else output[0]
    click.echo(
        f"Generating new user configuration file"
        f" '{output}'"
    )
    with fdp_session.FAIR(os.getcwd(), debug=debug) as fair_session:
        fair_session.make_starter_config(output)


@cli.command()
@click.option(
    "--config",
    help="Specify alternate location for generated config.yaml",
    default=fdp_com.local_user_config(os.getcwd()),
    shell_complete=complete_yamls
)
@click.option(
    "--using",
    help="Initialise the CLI system from an existing CLI global configuration file",
    default="",
    shell_complete=complete_yamls
)
@click.option(
    "--registry",
    help="Specify registry directory",
    default=fdp_svr.DEFAULT_REGISTRY_LOCATION,
    show_default=True
)
@click.option(
    "--ci/--standard",
    help="Run in testing mode for a CI system",
    default=False
)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.option("--export", help="Export the CLI configuration to a file", default="")
def init(
    config: str, debug: bool, using: str, registry: str, ci:bool, export: str = ""
    ) -> None:
    """Initialise repository in current location"""
    try:
        with fdp_session.FAIR(
            os.getcwd(), config, debug=debug, testing=ci
        ) as fair_session:
            _use_dict = {}
            if using:
                if not os.path.exists(using):
                    raise fdp_exc.FileNotFoundError(
                        f"Cannot load CLI configuration from file '{using}', "
                        "file does not exist."
                    )
                _use_dict = yaml.safe_load(open(using))
            fair_session.initialise(
                using=_use_dict,
                registry=registry,
                export_as=export
            )
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.command()
@click.option(
    "glob",
    "--global/--no-global",
    help="Also delete global FAIR-CLI directories",
    default=False
)
@click.option(
    "--yes/--no",
    help="Deletes the configurations specified without prompt",
    default=False
)
@click.option(
    "--data/--no-data",
    help="Also delete the local data directory",
    default=False
)
@click.option(
    "--all/--not-all",
    help="Remove all FAIR interfaces and registry",
    default=False
)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def purge(glob: bool, debug: bool, yes: bool, data: bool, all: bool) -> None:
    """Resets the repository deleting all local caches"""
    _purge = yes

    if all:
        all = click.confirm(
            "Are you sure you want to remove all FAIR components from this system?\n"
            "WARNING: This will also remove your local registry"
        )
    else:
        _purge = click.confirm(
            "Are you sure you want to reset FAIR tracking, "
            "this is not reversible?"
        )
        if data:
            data = click.confirm(
                "Are you sure you want to delete the local data directory?\n"
                "WARNING: Do not do this if you have a populated local registry"
            )
        if not _purge:
            return

    try:
        with fdp_session.FAIR(os.getcwd()) as fair_session:
            fair_session.purge(
                global_cfg=glob,
                local_cfg=_purge,
                clear_data=data,
                clear_all=all
            )
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.group()
def registry() -> None:
    """Commands relating to control of the local registry server"""
    pass

@registry.command()
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def uninstall(debug: bool):
    """Uninstall the local registry from the system"""
    _confirm = click.confirm(
        "Are you sure you want to remove the local registry and its components?",
        default=False
    )
    if not _confirm:
        return
    try:
        fdp_svr.uninstall_registry()
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)

@registry.command()
@click.option("--force/--no-force", help="Force a reinstall", default=False)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.option("--directory", help="Installation location", default=None)
def install(debug: bool, force: bool, directory: str):
    """Install the local registry on the system"""
    try:
        fdp_svr.install_registry(install_dir=directory, force=force)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@registry.command()
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def start(debug) -> None:
    """Start the local registry server"""
    try:
        fdp_session.FAIR(
            os.getcwd(), server_mode=fdp_svr.SwitchMode.USER_START
        )
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@registry.command()
@click.option("--force/--no-force", help="Force server stop", default=False)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def stop(force: bool, debug: bool) -> None:
    """Stop the local registry server"""
    _mode = (
        fdp_svr.SwitchMode.FORCE_STOP
        if force
        else fdp_svr.SwitchMode.USER_STOP
    )
    try:
        fdp_session.FAIR(os.getcwd(), server_mode=_mode)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.command()
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def log(debug: bool) -> None:
    """Show a full job history"""
    try:
        fdp_hist.show_history(os.getcwd())
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.command()
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.argument("job_id")
def view(job_id: str, debug: bool) -> None:
    """View log for a given job"""
    try:
        fdp_hist.show_job_log(os.getcwd(), job_id)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.command()
@click.argument("file_paths", type=click.Path(exists=True), nargs=-1)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def reset(file_paths: typing.List[str], debug: bool) -> None:
    """Removes jobs from staging"""
    pass


@cli.command()
@click.argument("job_ids", nargs=-1)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def add(job_ids: typing.List[str], debug: bool) -> None:
    """Add a job to staging"""
    pass


@cli.command()
@click.argument("job_ids", nargs=-1)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.option(
    "--cached/--not-cached",
    default=False,
    help="remove from tracking but do not delete from file system",
)
def rm(
    job_ids: typing.List[str], cached: bool = False, debug: bool = False
) -> None:
    """Removes jobs from system or just tracking"""
    pass


@cli.command()
@click.argument(
    "config",
    nargs=-1
)
@click.option(
    "--script",
    help="Specify a shell command to execute, this will be inserted into the working config",
    default=""
)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.option(
    "--ci/--no-ci",
    help="Calls run passively without executing any commands for a CI system",
    default=False
)


def run(config: str, script: str, debug: bool, ci: bool):
    """Initialises a job with the option to specify a bash command"""
    # Allow no config to be specified, if that is the case use default local
    if len(config) > 0:
        config = config[0]
    else:
        config = fdp_com.local_user_config(os.getcwd())
    _run_mode = fdp_run.CMD_MODE.RUN if not ci else fdp_run.CMD_MODE.PASS
    try:
        with fdp_session.FAIR(
            os.getcwd(), config, debug=debug, server_mode=fdp_svr.SwitchMode.CLI,
        ) as fair_session:
            _hash = fair_session.run_job(script, mode=_run_mode)
            if ci:
                click.echo(fdp_run.get_job_dir(_hash))
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)

@cli.group(invoke_without_command=True)
@click.option("--verbose/--no-verbose", "-v/")
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
@click.pass_context
def remote(ctx, verbose: bool = False, debug: bool = False):
    """List remotes if no additional command is provided"""
    if not ctx.invoked_subcommand:
        try:
            with fdp_session.FAIR(os.getcwd(), debug=debug) as fair_session:
                fair_session.list_remotes(verbose)
        except fdp_exc.FAIRCLIException as e:
            if debug:
                raise e
            e.err_print()
            if e.level.lower() == "error":
                sys.exit(e.exit_code)


@remote.command()
@click.argument("options", nargs=-1)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def add(options: typing.List[str], debug: bool) -> None:
    """Add a remote registry URL with option to give it a label if multiple
    remotes may be used.

    Parameters
    ----------
    options : typing.List[str]
        size 1 or 2 list containing either:
            - label, url
            - url
    """
    _url = options[1] if len(options) > 1 else options[0]
    _label = options[0] if len(options) > 1 else "origin"

    try:
        with fdp_session.FAIR(os.getcwd(), debug=debug) as fair_session:
            fair_session.add_remote(_url, _label)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@remote.command()
@click.argument("label")
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def remove(label: str, debug: bool) -> None:
    """Removes the specified remote from the remotes list

    Parameters
    ----------
    label : str
        label of remote to remove
    """
    try:
        with fdp_session.FAIR(os.getcwd(), debug=debug) as fair_session:
            fair_session.remove_remove(label)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@remote.command()
@click.argument("label")
@click.argument("url")
@click.pass_context
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def modify(ctx, label: str, url: str, debug: bool) -> None:
    """Modify a remote address"""
    try:
        with fdp_session.FAIR(os.getcwd(), debug=debug) as fair_session:
            fair_session.modify_remote(label, url)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.command()
@click.argument("remote", nargs=-1)
@click.option("--debug/--no-debug", help="Run in debug mode", default=False)
def push(remote: str, debug: bool):
    """Push data between the local and remote registry"""
    remote = 'origin' if len(remote) == 0 else remote[0]
    try:
        with fdp_session.FAIR(os.getcwd(), debug=debug) as fair_session:
            fair_session.push(remote)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)


@cli.group()
def config():
    """Configure user information"""
    pass


@config.command(name="user.name")
@click.argument("user_name")
def config_user(user_name: str) -> None:
    fdp_conf.set_user(os.getcwd(), user_name)


@config.command(name="user.email")
@click.argument("user_email")
def config_email(user_email: str) -> None:
    fdp_conf.set_email(os.getcwd(), user_email)


@cli.command()
@click.argument(
    "config",
    nargs=-1
)
@click.option('--debug/--no-debug')
def pull(config: str, debug: bool):
    """Update local registry from remotes and sources"""
    if len(config) > 0:
        config = config[0]
    else:
        config = fdp_com.local_user_config(os.getcwd())
    try:
        with fdp_session.FAIR(
            os.getcwd(),
            config,
            server_mode=fdp_svr.SwitchMode.CLI,
            debug=debug
        ) as fair:
            fair.run_job(mode=fdp_run.CMD_MODE.PULL)
    except fdp_exc.FAIRCLIException as e:
        if debug:
            raise e
        e.err_print()
        if e.level.lower() == "error":
            sys.exit(e.exit_code)

if __name__ in "__main__":
    cli()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Versioning
==========

Methods for handling semantic versioning and parsing version formats

Contents
========

Functions
---------
    parse_incrementer - parse version increment format

"""

__date__ = "2021-08-05"

import re
import typing
import semver

import fair.exceptions as fdp_exc
import fair.configuration as fdp_conf


BUMP_FUNCS = {
    "LATEST": None,
    "MINOR": "bump_minor",
    "MAJOR": "bump_major",
    "PATCH": "bump_patch",
    "BUILD": "bump_build",
    "PRERELEASE": "bump_prerelease"
}

DEFAULT_WRITE_VERSION = "PATCH"
DEFAULT_READ_VERSION = "LATEST"

def parse_incrementer(incrementer: str) -> str:
    """Convert an incrementer string in a config to the relevant bump function
    
    Parameters
    ----------
        incrementer : str
            config.yaml variable to describe how version increases are handled
    
    Returns
    -------
        str
            relevant member method of VersionInfo for 'bumping' the semantic version

    """
    # Sanity check to confirm all methods are still present in semver module
    for func in BUMP_FUNCS.values():
        if func and func not in dir(semver.VersionInfo):
            raise fdp_exc.InternalError(f"Unrecognised 'semver.VersionInfo' method '{func}'")

    for component in BUMP_FUNCS:
        try:
            if re.findall(r'\$\{\{\s*'+component+r'\s*\}\}', incrementer):
                return BUMP_FUNCS[component]
        except TypeError:
            raise fdp_exc.InternalError(
                f"Failed to parse incrementer '{incrementer}' expected string"
            )
    
    raise fdp_exc.UserConfigError(
        f"Unrecognised version incrementer variable '{incrementer}'"
    )


def undo_incrementer(incrementer: str) -> str:
    """Convert an incrementer string to just return the latest value
    
    Parameters
    ----------
        incrementer : str
            config.yaml variable to describe how version increases are handled for register
    
    Returns
    -------
        str
            correct value for version string for read

    """
    for component in BUMP_FUNCS:
        if re.findall(r'\$\{\{\s*'+component+r'\s*\}\}', incrementer):
            return re.sub(
                r'\$\{\{\s*'+component+r'\s*\}\}',
                '${{ LATEST }}',
                incrementer
            )

    return incrementer


def get_latest_version(results_list: typing.List = None) -> semver.VersionInfo:
    if not results_list:
        return semver.VersionInfo.parse("0.0.0")

    _versions = [
        semver.VersionInfo.parse(i['version']) for i in results_list
        if 'version' in i
    ]

    if not _versions:
        return semver.VersionInfo.parse("0.0.0")

    return max(_versions)


def get_correct_version(
    version: str,
    results_list: typing.List = None,
    free_write: bool = True
) -> semver.VersionInfo:

    _zero = semver.VersionInfo.parse("0.0.0")

    if results_list:
        _versions = [
            semver.VersionInfo.parse(i['version']) for i in results_list
            if 'version' in i
        ]
    else:
        _versions = []

    try:
        _bump_func = parse_incrementer(version)
        if free_write:
            _versions.append(_zero)

        if not _versions:
            raise fdp_exc.InternalError(
                f"Version parsing failed for version={version}, free_write={free_write}"
            )
        _max_ver = max(_versions)

        _new_version = getattr(_max_ver, _bump_func)() if _bump_func else _max_ver
    except fdp_exc.UserConfigError: # Not a command, try an exact version
        _new_version = semver.VersionInfo.parse(version)

    if _new_version in _versions and free_write:
        raise fdp_exc.UserConfigError(
            f"Trying to create existing version: {_new_version}"
        )
    elif _new_version not in _versions and not free_write:
        raise fdp_exc.UserConfigError(
            f"Trying to read non-existing version: {_new_version}"
        )
    elif _new_version == _zero:
        raise fdp_exc.UserConfigError(f'Trying to work with version {_zero}')

    return _new_version


def default_bump(version: semver.VersionInfo) -> semver.VersionInfo:
    """Perform default version bump

    For FAIR-CLI the default version increment is patch
    
    Parameters
    ----------
        version: semver.VersionInfo
    
    Returns
    -------
        new version
    """
    return getattr(version, BUMP_FUNCS[DEFAULT_WRITE_VERSION])()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Registry Requests
=================

Methods relating to connection with local and remote registries for the purpose
of synchronisation and push/pull.

Contents
========

    local_token   - retrieves the local API token
    post          - post to the registry RestAPI
    get           - retrieve from the registry RestAPI
    post_else_get - if an item already exists retrieve it, else create it

"""

__date__ = "2021-07-02"

import json
import os
import tempfile
import typing
import logging
import copy
import urllib.parse
import re
import requests
from requests.api import head
import simplejson.errors

import fair.common as fdp_com
import fair.exceptions as fdp_exc
import fair.utilities as fdp_util

SEARCH_KEYS = {
    "data_product": "name",
    "namespace": "name",
    "file_type": "extension",
    "storage_root": "root",
    "storage_location": "hash"
}

logger = logging.getLogger('FAIRDataPipeline.Requests')

def split_api_url(request_url: str, splitter: str = 'api') -> typing.Tuple[str]:
    """Split a request URL into endpoint and path

    Parameters
    ----------
    request_url : str
        URL to be split
    splitter : str, optional
        split point, by default 'api'

    Returns
    -------
    typing.Tuple[str]
        endpoint, path
    """
    _root, _path = request_url.split(f'{splitter}/')
    return f'{_root}{splitter}', _path


def local_token(registry_dir: str = None) -> str:
    """Read the local registry token from the relevant file"""
    registry_dir = registry_dir or fdp_com.registry_home()
    _local_token_file = os.path.join(registry_dir, "token")
    if not os.path.exists(_local_token_file):
        raise fdp_exc.FileNotFoundError(
            f"Failed to find local registry token, file '{_local_token_file}'"
            " does not exist.",
            hint="Try creating the file by manually starting the registry "
            "by running 'fair registry start'"
        )
    _file_lines = open(_local_token_file).readlines()

    if not _file_lines:
        raise fdp_exc.FileNotFoundError(
            f"Expected token in file {_local_token_file}, but file is empty"
        )

    return _file_lines[0].strip()


def _access(
    uri: str,
    method: str,
    obj_path: str = None,
    response_code: int = 200,
    token: str = None,
    headers: typing.Dict[str, typing.Any] = None,
    params: typing.Dict = None,
    data: typing.Dict = None,
    *args,
    **kwargs,
):
    if not headers:
        headers: typing.Dict[str, str] = {}

    if not params:
        params: typing.Dict[str, str] = {}
    
    if not data:
        data: typing.Dict[str, str] = {}

    if not token:
        token = local_token()

    # Make sure we have the right number of '/' in the components
    _uri = uri
    if _uri[-1] != "/":
        _uri += '/'

    _url = urllib.parse.urljoin(_uri, obj_path) if obj_path else uri

    if _url[-1] != "/":
        _url += "/"

    _headers = copy.deepcopy(headers)
    _headers["Authorization"] = f"token {token}"

    logger.debug("Sending request of type '%s': %s", method, _url)

    try:
        if method == 'get':
            _request = requests.get(
                _url, headers=_headers, params=params, *args, **kwargs
            )
        elif method == 'post':
            _request = requests.post(
                _url, headers=_headers, data=data, *args, **kwargs
            )
        else:
            _request = getattr(requests, method)(
                _url, headers=_headers, *args, **kwargs
            )
    except requests.exceptions.ConnectionError:
        raise fdp_exc.UnexpectedRegistryServerState(
            f"Failed to make registry API request '{_url}'",
            hint="Is this remote correct and the server running?"
        )

    _info = f'url = {_url}, '
    _info += f' parameters = {kwargs["params"]},' if 'params' in kwargs else ''
    _info += f' data = {kwargs["data"]}' if 'data' in kwargs else ''

    # Case of unrecognised object
    if _request.status_code == 404:
        raise fdp_exc.RegistryAPICallError(
            f"Attempt to access an unrecognised resource on registry "
            f"using method '{method}' and arguments: " + _info,
            error_code=404
        )

    # Case of unrecognised object

    if _request.status_code == 403:
        raise fdp_exc.RegistryAPICallError(
            f"Failed to run method '{method}' for url {_url}, "
            f"request forbidden",
            error_code=403
        )
    elif _request.status_code == 409:
        _searchable = uri if not obj_path else '/'.join(obj_path)
        raise fdp_exc.RegistryAPICallError(
            f"Cannot post object of type '{_searchable}' "
            f"using method '{method}' as it already exists."
            f"Arguments:\n"+_info,
            error_code=409,
        )

    try:
        _json_req = _request.json()
        _result = _json_req["results"] if "results" in _json_req else _json_req
    except (json.JSONDecodeError, simplejson.errors.JSONDecodeError):
        raise fdp_exc.RegistryAPICallError(
            f"Failed to retrieve JSON data from request to '{_url}'",
            error_code=_request.status_code
        )

    if _request.status_code != response_code:
        _info = ""
        if isinstance(_result, dict) and "detail" in _result:
            _info = _result["detail"]
        if not _info:
            _info = _result
        raise fdp_exc.RegistryAPICallError(
            f"Request failed with status code {_request.status_code}:"
            f" {_info}",
            error_code=_request.status_code,
        )
    return _result


def post(
    uri: str,
    obj_path: str,
    data: typing.Dict[str, typing.Any],
    headers: typing.Dict[str, typing.Any] = None,
    token: str = None
) -> typing.Dict:
    """Post an object to the registry

    Parameters
    ----------
    uri : str
        endpoint of the registry
    obj_path : str
        type of object to post
    data : typing.Dict[str, typing.Any]
        data for the object
    headers : typing.Dict[str, typing.Any], optional
        any additional headers for the request, by default None
    token : str, optional
        token for accessing the registry, by default None

    Returns
    -------
    typing.Dict
        resulting data returned from the request
    """
    if headers is None:
        headers = {}

    if not token:
        token = local_token()

    headers.update({"Content-Type": "application/json"})
    return _access(
        uri,
        "post",
        obj_path,
        201,
        headers=headers,
        data=json.dumps(data, cls=fdp_util.JSONDateTimeEncoder),
        token=token
    )


def url_get(url: str, *args, **kwargs) -> typing.Dict:
    """Send a URL only request and retrieve results

    Unlike 'get' this method is 'raw' in that there is no validation of
    components

    Parameters
    ----------
    url : str
        URL to send request to

    Returns
    -------
    typing.Dict
        results dictionary
    """
    return _access(url, "get", *args, **kwargs)


def get(
    uri: str,
    obj_path: str,
    headers: typing.Dict[str, typing.Any] = None,
    params: typing.Dict[str, typing.Any] = None,
    token: str = None
) -> typing.Dict:
    """Retrieve an object from the given registry

    Parameters
    ----------
    uri : str
        endpoint of the registry
    obj_path : str
        type of the object to fetch
    headers : typing.Dict[str, typing.Any], optional
        any additional headers for the request, by default None
    params : typing.Dict[str, typing.Any], optional
        search parameters for the object, by default None
    token : str, optional
        token for accessing the registry, by default None

    Returns
    -------
    typing.Dict
        returned data for the given request
    """
    logger.debug(
        "Retrieving object of type '%s' from registry at '%s' with parameters: %s",
        obj_path, uri, params
    )

    if not headers:
        headers = {}

    params = {} if not params else copy.deepcopy(params)

    if not token:
        token = local_token()

    if "namespace" in params and isinstance(params['namespace'], str):
        _namespaces = get(
            uri,
            'namespace',
            params = {SEARCH_KEYS['namespace']: params['namespace']}
        )

        if len(_namespaces) > 1:
            raise fdp_exc.UserConfigError(
                f"Multiple ({len(_namespaces)}) hits for namespace '{params['namespace']}'"
            )
        elif len(_namespaces) == 0:
            raise fdp_exc.UserConfigError(
                f"No hits for namespace '{params['namespace']}'"
            )

        _results = re.search(
            r'^' + uri + r'/?namespace/(\d+)/$',
            _namespaces[0]['url']
        )

        if not _results:
            raise fdp_exc.InternalError(
                "Failed to parse namespace identifiers"
            )

        params['namespace'] = int(_results.group(1))

    if "data_product" in params:
        _data_products = get(
            uri,
            'data_product',
            params = {
                SEARCH_KEYS['data_product']: params['data_product'],
                'namespace': params['namespace']
            }
        )

        _results = [re.search(
            r'^' + uri + r'/?data_product/(\d+)/$',
            _data_product['url']) for _data_product in _data_products
        ]

        _output = []
        del params['namespace']
        for data_product in _results:
            params['data_product'] = int(data_product.group(1))
            _output.extend(_access(
                uri,
                "get",
                obj_path,
                200,
                headers=headers,
                params=params,
                token=token
            ))
        return _output

    return _access(
        uri,
        "get",
        obj_path,
        200,
        headers=headers,
        params=params,
        token=token
    )


def post_else_get(
    uri: str,
    obj_path: str,
    data: typing.Dict[str, typing.Any],
    params: typing.Dict[str, typing.Any] = None,
    token: str = None
) -> str:
    """Post to the registry if an object does not exist else retrieve URL

    Parameters
    ----------
    uri : str
        endpoint of the registry
    obj_path : str
        object type to post
    data : typing.Dict[str, typing.Any]
        data for the object to be posted
    params : typing.Dict[str, typing.Any], optional
        parameters for searching if object exists, by default None
    token : str, optional
        token to access registry, by default None

    Returns
    -------
    str
        URL for object either retrieved or posted
    """
    if not params:
        params = {}

    if not token:
        token = local_token()

    try:
        _loc = post(uri, obj_path, data=data, token=token)
    except fdp_exc.RegistryAPICallError as e:
        # If the item is already in the registry then ignore the
        # conflict error and continue, else raise exception
        if e.error_code == 409:
            _loc = get(uri, obj_path, params=params)
        else:
            raise e

    if isinstance(_loc, list):
        if not _loc:
            raise fdp_exc.RegistryAPICallError(
                "Expected to receieve a URL location from registry post"
            )
        _loc = _loc[0]
    if isinstance(_loc, dict):
        _loc = _loc["url"]
    return _loc


def filter_object_dependencies(
    uri: str,
    obj_path: str,
    filter: typing.Dict[str, typing.Any]
    ) -> typing.List[str]:
    """Filter dependencies of an API object based on a set of conditions

    Parameters
    ----------
    uri : str
        endpoint of the registry
    object : str
        path of object type, e.g. 'code_run'
    filter : typing.Dict[str, typing.Any]
        list of filters to apply to listing

    Returns
    -------
    typing.List[str]
        list of object type paths
    """
    try:
        _actions = _access(uri, 'options', obj_path, 200)['actions']['POST']
    except KeyError:
        # No 'actions' key means no dependencies
        return []
    _fields: typing.List[str] = []

    for name, info in _actions.items():
        _filter_result: typing.List[bool] = []
        for filt, value in filter.items():
            # Some objects may not have the key
            if filt not in info:
                continue
            _filter_result.append(info[filt] == value)
        if all(_filter_result):
            _fields.append(name)

    return _fields


def get_filter_variables(uri: str, obj_path: str) -> typing.List[str]:
    """Retrieves a list of variables you can filter by for a given object

    Parameters
    ----------
    uri : str
        endpoint of registry
    obj_path : str
        type of object

    Returns
    -------
    typing.List[str]
        list of filterable fields
    """
    try:
        _filters = _access(uri, 'options', obj_path, 200)['filter_fields']
    except KeyError:
        # No 'filter_fields' key means no filters
        return []
    return [*_filters]


def get_writable_fields(
    uri: str,
    obj_path: str
    ) -> typing.List[str]:
    """Retrieve a list of writable fields for the given RestAPI object

    Parameters
    ----------
    uri : str
        endpoint of the registry
    object : str
        path of object type, e.g. 'code_run'

    Returns
    -------
    typing.List[str]
        list of object type paths
    """
    return filter_object_dependencies(
        uri, obj_path, {"read_only": False}
    )


def download_file(url: str, chunk_size: int = 8192) -> str:
    """Download a file from a given URL

    Parameters
    ----------
    url : str
        address of remote file
    chunk_size : int, optional
        chunk size for download, by default 8192

    Returns
    -------
    str
        path of downloaded temporary file
    """
    # Save the data to a temporary file so we can calculate the hash
    _file, _fname = tempfile.mkstemp()

    with requests.get(url, stream=True) as r_in:
        try:
            r_in.raise_for_status()
        except requests.HTTPError:
            raise fdp_exc.FileNotFoundError(
                f"Failed to download file from '{url}'"
                f" with status code {r_in.status_code}"
            )
        with os.fdopen(_file, 'wb') as in_f:
            for chunk in r_in.iter_content(chunk_size=chunk_size):
                in_f.write(chunk)

    return _fname


def get_dependency_listing(uri: str) -> typing.Dict:
    """Get complete listing of all objects and their registry based dependencies

    Parameters
    ----------
    uri : str
        endpoint of the registry

    Returns
    -------
    typing.Dict
        dictionary of object types and their registry based dependencies
    """

    _registry_objs = url_get(uri)

    return {
        obj: filter_object_dependencies(
            uri,
            obj,
            {
                "read_only": False,
                "type": "field",
                "local": True
            }
        ) for obj in _registry_objs
    }

def get_obj_type_from_url(request_url: str) -> str:
    """Retrieves the type of object from the given URL

    Parameters
    ----------
    request_url : str
        url to type check

    Returns
    -------
    str
        object type if recognised else empty string
    """
    _uri, _ = split_api_url(request_url)
    for obj_type in sorted([*url_get(_uri)], key=len, reverse=True):
        if obj_type in request_url:
            return obj_type
    return ""



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Common Paths
============

Functions and constant strings related to the location of directories
and files for executing a CLI session.


Contents
========

Members
-------
    USER_FAIR_DIR   - user FAIR directory
    FAIR_CLI_CONFIG - name of the FAIR-CLI configuration file
    FAIR_FOLDER     - name for FAIR local repository directory

Functions
-------
    registry_home       - returns the location of the local data registry
    find_fair_root      - returns the closest '.fair' directory in the upper hierarchy
    find_git_root       - returns the closest '.git' directory
    staging_cache       - returns the current repository staging cache directory
    default_data_dir    - returns the default data store
    local_fdpconfig     - returns path of FAIR-CLI local repository config
    local_user_config   - returns the path of the user config in the given folder
    default_jobs_dir    - returns the default jobs folder
    global_config_dir   - returns the FAIR-CLI global config directory
    global_fdpconfig    - returns path of FAIR-CLI global config
    session_cache_dir   - returns location of session cache folder

"""
__date__ = "2021-06-28"

import os
import pathlib
import yaml
import enum
import git

import fair.exceptions as fdp_exc

USER_FAIR_DIR = os.path.join(pathlib.Path.home(), ".fair")
FAIR_CLI_CONFIG = "cli-config.yaml"
USER_CONFIG_FILE = "config.yaml"
FAIR_FOLDER = ".fair"
JOBS_DIR = "jobs"

class CMD_MODE(enum.Enum):
    RUN = 1,
    PULL = 2,
    PUSH = 3,
    PASS = 4


def registry_home() -> str:
    _glob_conf = yaml.safe_load(open(global_fdpconfig()))
    if 'registries' not in _glob_conf:
        raise fdp_exc.CLIConfigurationError(
            "Expected key 'registries' in global CLI configuration"
        )

    if 'local' not in _glob_conf['registries']:
        raise fdp_exc.CLIConfigurationError(
            "Expected 'local' registry in global CLI configuration registries"
        )

    if 'directory' not in _glob_conf['registries']['local']:
        raise fdp_exc.CLIConfigurationError(
            'Expected directory of local registry in global CLI configuration'
        )

    return _glob_conf['registries']['local']['directory']


def find_fair_root(start_directory: str = os.getcwd()) -> str:
    """Locate the .fair folder within the current hierarchy

    Parameters
    ----------

    start_directory : str, optional
        starting point for local FAIR folder search

    Returns
    -------
    str
        absolute path of the .fair folder
    """
    _current_dir = os.path.abspath(start_directory)

    while _current_dir:
        # If the home directory has been reached then abort as upper file system
        # is outside user area, also we do not want to return global FAIR folder
        if str(_current_dir) == str(pathlib.Path().home()):
            return ""

        if os.path.exists(os.path.join(_current_dir, FAIR_FOLDER)):
            return _current_dir
        _current_dir, _directory = os.path.split(_current_dir)

        # If there is no directory component this means the top of the file
        # system has been reached
        if not _directory:
            return ""            


def staging_cache(user_loc: str) -> str:
    """Location of staging cache for the given repository"""
    return os.path.abspath(
        os.path.join(find_fair_root(user_loc), FAIR_FOLDER, "staging")
    )


def default_data_dir(location: str = 'local') -> str:
    """Location of the default data store"""
    if not os.path.exists(global_fdpconfig()):
        raise fdp_exc.InternalError(
            f"Failed to read CLI global config file '{global_fdpconfig()}'"
        )
    _glob_conf = yaml.safe_load(open(global_fdpconfig()))
    if 'data_store' in _glob_conf['registries'][location]:
        return _glob_conf['registries'][location]['data_store']
    if location == 'local':
        return os.path.join(USER_FAIR_DIR, f"data{os.path.sep}")
    else:
        raise fdp_exc.UserConfigError('Cannot guess remote data store location')


def local_fdpconfig(user_loc: str = os.getcwd()) -> str:
    """Location of the FAIR-CLI configuration file for the given repository"""
    return os.path.join(find_fair_root(user_loc), FAIR_FOLDER, FAIR_CLI_CONFIG)


def local_user_config(user_loc: str = os.getcwd()) -> str:
    """Location of the default user configuration file for the given repository"""
    return os.path.join(find_fair_root(user_loc), USER_CONFIG_FILE)


def default_jobs_dir() -> str:
    """Default location to place job outputs"""
    return os.path.join(default_data_dir(), JOBS_DIR)


def global_config_dir() -> str:
    """Directory of global CLI configuration"""
    return os.path.join(USER_FAIR_DIR, "cli")


def session_cache_dir() -> str:
    """Location of run files used to determine if server is being used"""
    return os.path.join(global_config_dir(), "sessions")


def global_fdpconfig() -> str:
    """Location of global CLI configuration"""
    return os.path.join(global_config_dir(), FAIR_CLI_CONFIG)


def find_git_root(start_directory: str = os.getcwd()) -> str:
    """Locate the .git folder within the current hierarchy

    Parameters
    ----------

    start_directory : str, optional
        starting point for local git folder search

    Returns
    -------
    str
        absolute path of the .git folder
    """
    try:
        _repository = git.Repo(
            start_directory,
            search_parent_directories=True
        )
    except git.InvalidGitRepositoryError:
        raise fdp_exc.UserConfigError(
            "Failed to retrieve git repository for current configuration"
            f" in location '{start_directory}'"
        )
    return _repository.git.rev_parse("--show-toplevel").strip()
